#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################
class IncidentsController < ApplicationController
  helper Redborder::ModulesHelper

  include ApplicationHelper

  include IncidentsHelper

  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_incident, only: [:edit, :edit_description, :update, :destroy, :show, :next_phase, :previous_phase, :close]
  before_action :set_priorities, only: [:index, :show, :search, :update]
  before_action :set_statuses, only: [:index, :show, :search, :update]
  before_action :assign_first_user_and_open!, only: [:show]
  before_action :set_users, only: [:show]

  ## ======================================================================
  # => Actions
  ## ======================================================================

  #
  # => Render all incidents in a table
  #
  def index
    @incidents = Incident.where(visible: true).order(created_at: :desc)

    @priorities = Priority.all
    @incident_statuses = IncidentStatus.all

    # Paginate results
    @page = params[:page]
    @incidents = @incidents.paginate(page: params[:page],
                               per_page: current_user.per_page_count)
  end

  #
  # => Render the form to create a new incident
  #
  def new
    @incident = Incident.new(priority_id: Priority.first.id)
    @incident.incident_status = IncidentStatus.find_by(name: IncidentStatus::NEW)
  end

  #
  # => Receive the data to create a new incident. Try to create a new incident
  #    and redirect to incidents_path if it's ok
  #
  def create
    @incident = Incident.new(incident_params)

    respond_to do |format|
      if @incident.save
        format.html do
          redirect_to incidents_path,
                      notice: "Incident was successfully made"
        end
        format.js { flash[:notice] = "Incident was successfully made" }
        format.json { head :no_content }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @incident.errors.full_messages.map do
            |msg| "<li>#{msg}</li>"
          end.join
        end
        format.json do
          render json: @incident.errors, status: :unprocessable_entity
        end
      end
    end
  end

  #
  # => Render the form to edit an incident
  #
  def edit
    respond_to do |format|
      format.js
    end
  end

  #
  # => Render the form to edit an incident
  #
  def edit_description
    respond_to do |format|
      format.js
    end
  end

  #
  # => Receive the data to update a new incident. Try to update a new incident
  #    and redirect to incident_path if it's ok
  #
  def update
    @page = params[:page] || 1

    respond_to do |format|
      if @incident.update(incident_params)
        format.html do
          notice_message = "Incident was successfully updated"

          if request.referer && request.referer.include?(incident_path(@incident))
            redirect_to incident_path(@incident), notice: notice_message
          else
            redirect_to incidents_path(page: @page), notice: notice_message
          end
        end
        format.js { flash[:notice] = "Incident was successfully updated" }
        format.json { render json: @incident }
      else
        format.html { render action: 'show' }
        format.js do
          flash[:error] = @incident.errors.full_messages.map do |msg|
            "<li>#{msg}</li>"
          end.join
        end
        format.json { render json: @incident.errors, status: :unprocessable_entity }
      end
    end
  end

  #
  # => Destroy an incident
  #
  def destroy
    @incident.destroy
    respond_to do |format|
      format.html { redirect_to incidents_path }
      format.js
      format.json { head :no_content }
    end
  end

  def next_phase
    @current_phase = @incident.playbook.phases.current || @incident.playbook.phases.order(:id).first
    @current_phase.set_as_current if @current_phase

    if @current_phase && @next_phase = @current_phase.next_phase
      @next_phase.set_as_current
    end

    respond_to do |format|
      format.html { redirect_to incident_path(@incident, anchor: "response") }
      format.js
    end
  end

  def previous_phase
    @current_phase = @incident.playbook.phases.current || @incident.playbook.phases.order(:id).first
    @current_phase.set_as_current if @current_phase

    if @current_phase && @previous_phase = @current_phase.previous_phase
      @previous_phase.set_as_current
    end

    respond_to do |format|
      format.html { redirect_to incident_path(@incident, anchor: "response") }
      format.js
    end
  end

  def close
    respond_to do |format|
      if @incident.update(incident_status: IncidentStatus.find_by(name: IncidentStatus::CLOSED))
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), notice: "Incident status was successfully updated"
        end
        format.js { flash.now[:notice] = "Incident status was successfully updated" }
        format.json { render json: @incident }
      else
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), alert: "Error updating incident status"
        end
        format.js { flash.now[:error] = @incident.errors.full_messages.join(", ") }
        format.json { render json: @incident.errors, status: :unprocessable_entity }
      end
    end
  end

  def bulk_delete
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)
    if incident_ids.any?
      Incident.where(id: incident_ids).destroy_all
      flash[:notice] = "Incidents were successfully deleted"
    else
      flash[:alert] = "No incidents selected for deletion"
    end

    redirect_to incidents_path(page: page)
  end

  def bulk_update_status
    page = params[:page] || 1
    incident_ids = (params[:incident_ids] || []).reject(&:blank?).map(&:to_i)
    status = IncidentStatus.find_by(name: params[:status])
    status_id = status.id if status
    incidents = Incident.where(id: incident_ids)

    if status && incidents.update_all(incident_status_id: status_id)
      flash[:notice] = "Incidents were successfully updated"
    else
      flash[:alert] = "Failed to update incidents"
    end

    redirect_to incidents_path(page: page)
  end

  #
  # Search on incidents
  #
  # TODO: split this in two methods: fast_search and search
  def search
    @page = params[:page] || 1


    # Search box on the top
    if params[:fast_search].present?
      filter = ['( LOWER(name) LIKE :search or ' \
                      'LOWER(description) LIKE :search or LOWER(source) LIKE :search)',
                      search: "%#{params[:fast_search].try(:downcase)}%"]
      @incidents = Incident.where(visible: true).where(filter).order(created_at: :desc)
                   .paginate(page: @page, per_page: current_user.per_page_count)

    else
      # Search form
      query_params = []
      values = {}

      if params[:search][:priority].present?
        query_params << 'priority_id = :priority'
        values[:priority] = params[:search][:priority].downcase
      end

      if params[:search][:source].present?
        query_params << 'LOWER(source) = :source'
        values[:source] = params[:search][:source].downcase
      end

      if params[:search][:username].present?
        query_params << 'user_id = :user'
        values[:user] = params[:search][:username]
      end

      if params[:search][:status].present?
        query_params << 'incident_status_id = :status'
        values[:status] = params[:search][:status].downcase
      end

      @incidents = Incident.where(visible: true).where(query_params.join(' AND '), values).order(created_at: :desc)
                .paginate(page: @page, per_page: current_user.per_page_count)
    end
  end

  def search_user
    incident = Incident.find(params[:id])
    query = params[:q]

    users = User.where("name LIKE ?", "%#{query}%")

    # Filter users that are not already in the incident
    filtered_users = users
      .where(id: current_user.users_to_assign.pluck(:id))
      .where.not(id: incident.users.pluck(:id))

    render json: filtered_users.map { |user| { id: user.id, name: user.name } }
  end

  def show
    @actions = gather_all_actions(@incident)
    @worklog_data = sort_worklog_data(gather_all_worklog_data(@incident), :desc)
    @actions_taken_data = fetch_actions_taken_data(@incident)
    @actions_taken_data = sort_worklog_data(@actions_taken_data, :desc)

    search_params = params[:detection_search] || {}
    search_params = search_params.reject { |key, value| value.blank? }
    @elements = get_incident_druid_data(@incident, search_params)

    @detection_unique_sources = @elements.map { |log| log[:source] }.uniq
    @detection_unique_priorities = @elements.map { |log| log[:priority] }.uniq

    @incident_nodes = []
    @incident_links_bidirectional = []

    @elements.each do |element|
      # Assets
      if element[:assets] && !element[:assets].empty?
        element[:assets].each do |asset_element|
          asset_element.each do |key, value|
           @incident_nodes.push({ name: key, type: value, image: get_asset_image(value), asset: true })
          end
        end
      end

      # Observables
      if element[:observables] && !element[:observables].empty?
        element[:observables].each do |observable_element|
          observable_element.each do |key, value|
            @incident_nodes.push({ name: key, type: value, image: get_observable_image(value), observable: true })
           end
        end
      end
    end

    # Sort elements by count
    grouped_assets = @incident_nodes.select { |node| node[:asset] == true }.group_by { |e| e }
    @assets = grouped_assets.map { |key, value| { key: key, count: value.size } }.sort_by { |e| -e[:count] }

    grouped_observables = @incident_nodes.select { |node| node[:observable] == true }.group_by { |e| e }
    @observables = grouped_observables.map { |key, value| { key: key, count: value.size } }.sort_by { |e| -e[:count] }

    @incident_nodes = @incident_nodes.uniq

    @incident_nodes.each do |node|
      related_objects = []
      @elements.each do |element|
        objects = (element[:assets] + element[:observables]).map(&:keys).flatten
        if objects.include? node[:name]
          related_objects += objects
        end
      end

      # Make the link
      related_objects.uniq.each do |object|
        related_node = @incident_nodes.find { |hash| hash[:name] == object }

        next unless related_node

        @incident_links_bidirectional.push({source: @incident_nodes.find_index(node), target: @incident_nodes.find_index(related_node) ,value: 1})
      end
    end

    @incident_links = []

    @incident_links_bidirectional.each do |link|
      unless @incident_links.any? { |existing_link| (existing_link[:source] == link[:target] && existing_link[:target] == link[:source]) }
        @incident_links << link
      end
    end


    # Indicators
    @indicators = Hash.new(0)

    grouped_elements = @elements.group_by { |event| [event[:msg], event[:source], event[:sig_id]] }
    @indicators = grouped_elements.map do |(msg, source, sig_id), entries|
      {
        source: source,
        msg: msg,
        count: entries.size,
        sig_id: sig_id
      }
    end

    @indicators = @indicators.sort_by { |entry| -entry[:count] }

    # Get Mitre techniques / subtechniques
    @tactics = []
    @techniques = []
    @subtechniques = []

    @indicators.each do |indicator|
      next if indicator[:sig_id] == "N/A"

      sid = indicator[:sig_id]
      tactics = Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_tactic = ?", [sid], true)
      @tactics.concat(tactics).uniq! { |tactic| tactic.tactic_id }

      techniques = Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_technique = ?", [sid], true)
      @techniques.concat(techniques).uniq! { |technique| technique.technique_id }

      subtechniques = Mitre.where("rule_ids && ARRAY[?]::integer[] AND is_subtechnique = ?", [sid], true)
      @subtechniques.concat(subtechniques).uniq! { |subtechnique| subtechnique.subtechnique_id }
    end

    # Define current phase
    unless @incident.playbook.phases.current
      @incident.playbook.phases.order(:id).first.set_as_current
    end

    @current_phase = @incident.playbook.phases.current

    # Paginate results
    @page = params[:page] || 1
    @elements = @elements.paginate(page: params[:page], per_page: current_user.per_page_count)
  end

  def new_user
    @incident = Incident.find(params[:id])
    @users = current_user.users_to_assign
    respond_to do |format|
      format.js
    end
  end

  # TODO: Add checks permissions
  # Assign a user to the incident
  def assign_user
    incident = Incident.find(params[:id])
    user = User.find(params[:user_id])
    incident.users << user

    gravatar_uri = get_gravatar_uri(user.email)

    if incident.save
      notification = Notification.create(
        user_id: user.id,
        comment: I18n.t('notifications.incident.assigned', incident_id: incident.id),
        incident_id: incident.id
      )
      notification.save
    end

    render json: { id: incident.id, user_id: user.id, user_name: user.name, user_email: user.email, gravatar_uri: gravatar_uri }
  end

  # TODO: Add checks permissions
  def unassign_user
    incident = Incident.find(params[:id])
    user = User.find(params[:user_id])

    if incident.users.delete(user)
      respond_to do |format|
        format.html { redirect_to incident_path(incident), notice: 'User successfully removed from the incident.' }
        format.js
      end
      if incident.save
        notification = Notification.create(
          user_id: user.id,
          comment: I18n.t('notifications.incident.unassigned', incident_id: incident.id),
          incident_id: incident.id
        )
        notification.save
      end
    else
      respond_to do |format|
        format.html { redirect_to incident_path(incident), alert: 'Failed to remove user from the incident.' }
        format.js
      end
    end
  end

  def sidebar_observable_content
    @observable = params[:observable]
    @observable_links, @incident_nodes = related_observables(params[:incident_id], params[:observable][:name])
    @events = related_events(params[:incident_id], params[:observable][:name])
    
    render partial: 'sidebar_observable_content', locals: { observable: @observable, observable_links: @observable_links, incident_nodes: @incident_nodes, events: @events }
  end

  def sidebar_content
    @incident = Incident.find(params[:id])

    @elements = get_incident_druid_data(@incident, {})
    @assets = Hash.new(0)

    @elements.each do |element|
      # Assets
      if element[:assets] && !element[:assets].empty?
        element[:assets].each do |asset|
          asset.each do |key, value|
            @assets[key] += 1
           end
        end
      end
    end

    # Sort elements by count
    @assets = @assets.sort_by { |asset, count| -count }

    Rails.logger.error "@Assets is #{@assets}"
    render partial: 'sidebar_content', locals: { incident: @incident, assets: @assets }
  end

  def new_note
    @incident = Incident.find(params[:id])
    @note = Note.new

    respond_to do |format|
      format.js
    end
  end

  def new_response_action_note
    @incident = Incident.find(params[:id])
    @task = Task.find(params[:task_id])
    @note = Note.new

    respond_to do |format|
      format.js
    end
  end

  def new_response_action_execute
    @incident = Incident.find(params[:id])
    @task = Task.find(params[:task_id])

    Rails.logger.info "Task is #{@task.id}"

    @action = nil

    if @task.execution
      actions = IntegrationService.get_actions_by_integration(@task.integration_service)
      @action = actions[@task.action] rescue nil
    end

    Rails.logger.info "Action is #{@action}"

    respond_to do |format|
      format.js
    end
  end

  def create_note
    @incident = Incident.find(params[:id])
    @note = Note.new(content: note_params[:content])
    @note.user = current_user
    @note.incident = @incident

    respond_to do |format|
      if @note.save
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "Note was successfully created"
        end
        format.js { flash[:notice] = "Note was successfully created" }
        format.json { head :no_content }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @note.errors.full_messages.join(", ")
        end
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_response_action_note
    @incident = Incident.find(params[:id])
    @task = Task.find(note_params[:task_id])
    @phase = @task.phase

    # TODO: make this atomic
    @response_action = ResponseAction.new
    @response_action.task = @task
    @response_action.user = current_user
    @response_action.save

    @note = Note.new(content: note_params[:content])
    @note.user = current_user
    @note.response_action =  @response_action

    respond_to do |format|
      if @note.save
        format.html do
          redirect_to incident_path(@incident, anchor: "response"), notice: "Note was successfully created"
        end
        format.js { flash[:notice] = "Note was successfully created" }
        format.json { head :no_content }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @note.errors.full_messages.join(", ")
        end
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  def create_response_action_execute
    @incident = Incident.find(params[:id])
    @task = Task.find(params[:task_id])

    @action = {}

    if @task.execution
      actions = IntegrationService.get_actions_by_integration(@task.integration_service)
      @action = actions[@task.action] rescue {}
    end

    action_params = {}
    @action.keys.each { |key| action_params[key] = params[key] unless params[key].empty? || params[key].nil? }

    # TODO: make this atomic
    @response_action = ResponseAction.new
    @response_action.task = @task
    @response_action.user = current_user
    @response_action.save

    Delayed::Job.enqueue(
      Redborder::Jobs::ResponseActionJob.new(true, current_user.id, @response_action.id, @task.integration_service, @task.action, action_params),
      priority: 10,
      run_at: Time.now + 2.seconds
    )

     #stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{@response_action.id}%").first
   
    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "response"), notice: "Action was successfully queued for execution"
      end
      format.js { flash[:notice] = "Action was successfully queued for execution" }
      format.json { head :no_content }
    end
  end

  def edit_note
    @incident = Incident.find(params[:id])
    @note = Note.find(params[:note_id])

    respond_to do |format|
      format.js
    end
  end

  # TODO: only creator of node can update his note
  def update_note
    @incident = Incident.find(params[:id])
    @note = Note.find(params[:note_id])

    respond_to do |format|
      if @note.update(content: note_params[:content])
        format.html do
          redirect_to incident_path(@incident, anchor: "worklog"), notice: "Note was successfully created"
        end
        format.js { flash[:notice] = "Note was successfully updated" }
        format.json { render json: @incident }
      else
        format.html { render action: 'new' }
        format.js do
          flash[:error] = @note.errors.full_messages.join(", ")
        end
        format.json { render json: @note.errors, status: :unprocessable_entity }
      end
    end
  end

  # TODO: Check permissions for this with a before_action or whatever
  def destroy_note
    @incident = Incident.find(params[:id])
    @note = Note.find(params[:note_id])

    @note.destroy

    respond_to do |format|
      format.html do
        redirect_to incident_path(@incident, anchor: "worklog"), notice: "Note was successfully destroyed"
      end
      format.js
      format.json { head :no_content }
    end
  end

  def worklog_search
    @incident = Incident.find(params[:id])
    search_params = params[:worklog_search] || {}

    type = search_params[:type].presence || 'All'
    user_id = search_params[:user]
    order = search_params[:order] == 'Newest' ? :desc : :asc

    @worklog_data = fetch_worklog_data(type, @incident)
    @worklog_data = filter_worklog_data_by_user(@worklog_data, user_id) if user_id.present?
    @worklog_data = sort_worklog_data(@worklog_data, order)

    @worklog_data
  end

  def detection_search
    @page = params[:page] || 1
    @incident = Incident.find(params[:id])

    search_params = params[:detection_search] || {}
    search_params = search_params.reject { |key, value| value.blank? }
    @elements = get_incident_druid_data(@incident, search_params)

    @detection_unique_sources = @elements.map { |log| log[:source] }.uniq
    @detection_unique_priorities = @elements.map { |log| log[:priority] }.uniq

    # Paginate results
    @page = params[:page] || 1
    @elements = @elements.paginate(page: params[:page], per_page: current_user.per_page_count)
  end

  private

  #
  # => Init the incident if the user has permissions
  #
  def set_incident
    @incident = Incident.find(params[:id])
  end

  def set_priorities
    @priorities = Priority.all.order(:id)
  end

  def set_statuses
    @incident_statuses = IncidentStatus.all
  end

  # Those are the users the current_user can manage to add into the incident
  def set_users
    @users = current_user.users_to_share
  end

  # Assign firs user that visits a new incident and open it
  def assign_first_user_and_open!
    if @incident.users.empty? && current_user && @incident.new?
      @incident.users.push(current_user)
      if @incident.save
        notification = Notification.create(
          user_id: current_user.id,
          comment: I18n.t('notifications.incident.assigned', incident_id: @incident.id),
          incident_id: @incident.id
        )
        notification.save
        unless @incident.open!
          Rails.logger.error "Failed to open the incident: #{@incident.errors.full_messages.join(', ')}"
        end
      else
        Rails.logger.error "Failed to save the incident: #{@incident.errors.full_messages.join(', ')}"
      end
    end
  end

  def gather_all_actions(incident)
    response_actions = incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions)
    all_actions = (response_actions + response_actions).sort_by(&:created_at)
    all_actions
  end

  def gather_all_worklog_data(incident)
    incident_changes = gather_incident_audit_worklog_data(incident)
    notes = gather_response_action_notes_worklog_data(incident) + gather_notes_worklog_data(incident)
    executions = gather_response_action_executions_worklog_data(incident)
    all_worklog_data = (incident_changes + notes + executions)
  end

  def gather_incident_audit_worklog_data(incident)
    incident.audits.map do |audit|
      formatted_content = audit.comment

      if audit.audited_changes['incident_status_id']
        # Obtener los valores antiguos y nuevos del estado
        old_status_id, new_status_id = audit.audited_changes['incident_status_id']

        # Buscar los nombres de los estados antiguos y nuevos
        old_status = IncidentStatus.find_by(id: old_status_id).try(:name) ? IncidentStatus.find_by(id: old_status_id).name : old_status_id
        new_status = IncidentStatus.find_by(id: new_status_id).try(:name) ? IncidentStatus.find_by(id: new_status_id).name : new_status_id

        # Reemplazar el texto en el comentario
        formatted_content = formatted_content.gsub(/incident_status_id from '(\d+)' to '(\d+)'/) do
          "status from '#{old_status}' to '#{new_status}'"
        end
      end

      {
        action: audit.action,
        user: audit.user,
        created_at: audit.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: formatted_content
      }
    end
  end

  def gather_response_action_executions_worklog_data(incident)
    incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map do |response_action|
      if response_action.notes.any?
        []
      else
        stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{response_action.id}%").first
        stored_job_id = stored_job.present? ? stored_job.id : nil
        [{
          action: "Executed Action",
          user: response_action.user,
          created_at: response_action.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
          source: "[RESPONSE TASK] #{response_action.task.name}",
          stored_job_id: stored_job_id
        }]
      end
    end.compact
  end

  def gather_response_action_notes_worklog_data(incident)
    return incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map(&:notes).map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        source: "[RESPONSE TASK] #{note.response_action.task.name}",
        content: note.content
      }
    end
  end

  def gather_notes_worklog_data(incident)
    return incident.notes.map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: note.content
      }
    end
  end

  def fetch_worklog_data(type, incident)
    case type
    when 'Notes'
      (gather_response_action_notes_worklog_data(incident) +  gather_notes_worklog_data(incident))
    when 'Incident Changes'
      gather_incident_audit_worklog_data(incident)
    when 'Response Actions'
      gather_response_action_notes_worklog_data(incident) + gather_response_action_executions_worklog_data(incident)
    else
      gather_all_worklog_data(incident)
    end
  end

  def format_actions_taken_note(note)
    action = note[:action]
    if note[:action] == "Create"
      action = "Added a note for"
    end
    user = note[:user]
    source = note[:source].sub("[RESPONSE TASK] ", "")
    content = note[:content]
    created_at = note[:created_at]

    { action: action, user: user, created_at: created_at, source: source, content: content }
  end

  def format_actions_taken_execution(execution)    
    action = execution[:action]
    if execution[:action] == "Executed Action"
      action = "Executed Action for "
    end
    user = execution[:user]
    source = execution[:source].sub("[RESPONSE TASK] ", "")
    # content = execution[:content]
    created_at = execution[:created_at]
    stored_job_id = execution[:stored_job_id]

    { action: action, user: user, created_at: created_at, source: source, stored_job_id: stored_job_id }
  end

  def fetch_actions_taken_data(incident)
    data = []
    notes = gather_response_action_notes_worklog_data(incident)
    notes.each do |note|
      data.push(format_actions_taken_note(note))
    end
    
    executions = gather_response_action_executions_worklog_data(incident)

    executions.each do |execution|
      data.push(format_actions_taken_execution(execution))
    end

    data
  end

  def filter_worklog_data_by_user(worklog_data, user_id)
    user = User.find_by(id: user_id)
    return worklog_data unless user

    worklog_data.select { |item| item[:user].id == user.id }
  end

  def sort_worklog_data(worklog_data, order)
    date_format = "%B %d, %Y %I:%M:%S %p"

    converted_data = worklog_data.map do |item|
      item[:created_at_datetime] = DateTime.strptime(item[:created_at], date_format)
      item
    end

    sorted_data = converted_data.sort_by { |item| item[:created_at_datetime] }

    sorted_data.reverse! if order == :desc

    sorted_data.each { |item| item.delete(:created_at_datetime) }

    sorted_data
  end

  def get_incident_druid_data(incident, search_params)
    data = []

    @objects = RedborderObject.where(user_id: current_user.id)

    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }
   
    incident_modules.each do |mod| 
      source = Redborder::DruidSource.get(source: mod.first)
      filter = Redborder::Filter.new(source, {})
      filter.add(:incident_uuid, incident.uuid)

      start_time = incident.created_at - 1.hour
      end_time = start_time + 1.day

      # group_by: @source.default_columns_raw,
      data += source.where({
        query_type: :group_by,
        range: :custom,
        start_time: start_time,
        end_time: end_time,
        granularity: :minute,
        aggregation: source.default_aggregation,
        group_by: source.default_columns_incidents,
        search: filter.get
      }).map do |t|
        url_filters = {}
        url_filters[:start_time] = start_time
        url_filters[:end_time] = end_time

        priority = "N/A"
        if t.methods.include?(:priority)
          priority = t.send(:priority) || "N/A"
        end

        source_name = mod.last[:incident_source_name] rescue "N/A"
        
        module_name = mod.first

        timestamp = t.send(:timestamp) 

        msg = t.methods.include?(:msg) ? t.send(:msg) : nil

         # Observables are define in the columns of each module
        observables = []
        assets = []
        mod.last[:columns].select { |column, cfg| cfg[:observable_type] }.each do |column, cfg|
          asset = nil
          url_filters[column] = t.send(column)

          # We check if this is an asset
          if cfg[:object_type] && t.methods.include?(column) && t.send(column)
            asset = RedborderObject.find_by(value: t.send(column), type: cfg[:object_type], user: current_user) 
          end

          if asset
            assets.push({ t.send(column) => cfg[:object_type] }) if t.methods.include?(column) && t.send(column)
          else
            observables.push( { t.send(column) => cfg[:observable_type] }) if t.methods.include?(column) && t.send(column)
          end
        end

        url_filters["incident_uuid"] = incident.uuid
       
        { timestamp: timestamp, priority: priority, msg: msg, source: source_name, assets: assets, observables: observables, module_name: module_name, url_filters: url_filters }
      end
    end
    
    # Filter by search
    if search_params && search_params[:source]
      data = data.select { |element| element[:source] == search_params[:source] }
    end

    if search_params && search_params[:priority]
      data = data.select { |element| element[:priority] == search_params[:priority] }
    end

    data
  end

  # TODO: each module should define this
  def get_module_source_name(m)
    case m
    when :ips
      "redBorder Intrusion"
    when :vault
      "redBorder Vault"
    else
      m
    end
  end

  def get_asset_image(type)
    view_context.asset_path_from_controller("icons/assets/#{type}.svg")
  end

  def get_observable_image(type)
    view_context.asset_path_from_controller("icons/observables/#{type}.svg")
  end

  #
  # => Return a hash with the params to create/update an incident
  #
  def incident_params
    params
      .require(:incident)
      .permit(:name, :source, :description, :status, :user,
              :priority_id, :incident_status_id, :page, :search, :fast_search)
  end

  def note_params
    params.require(:note).permit(:id, :content, :task_id)
  end

  def related_observables(incident_id, observable_name)
    incident = Incident.find(incident_id)
    elements = get_incident_druid_data(incident, {})
    incident_nodes = []
    incident_links_bidirectional = []
  
    # Extract observable nodes from the incident data
    elements.each do |element|
      if element[:observables] && !element[:observables].empty?
        element[:observables].each do |observable_element|
          observable_element.each do |key, value|
            incident_nodes.push({ name: key, type: value, image: get_observable_image(value), observable: true })
          end
        end
      end
    end
  
    # Remove duplicate nodes
    incident_nodes = incident_nodes.uniq
  
    # Build bidirectional links between nodes
    incident_nodes.each do |node|
      related_objects = []
  
      elements.each do |element|
        objects = (element[:assets] + element[:observables]).map(&:keys).flatten
        if objects.include? node[:name]
          related_objects += objects
        end        
      end
  
      related_objects.uniq.each do |object|
        related_node = incident_nodes.find { |hash| hash[:name] == object }
        next unless related_node
  
        incident_links_bidirectional.push({
          source: incident_nodes.find_index(node),
          target: incident_nodes.find_index(related_node),
          value: 1
        })
      end
    end
  
    # Ensure that each link is unique and does not duplicate the reverse direction
    incident_links = []
  
    incident_links_bidirectional.each do |link|
      unless incident_links.any? { |existing_link| (existing_link[:source] == link[:target] && existing_link[:target] == link[:source]) }
        incident_links << link
      end
    end

    # Filter links involving the specified observable
    observable = incident_nodes.find { |node| node[:name] == observable_name }
  
    observable_links = incident_links.select do |link|
      link[:source] == incident_nodes.find_index(observable) || link[:target] == incident_nodes.find_index(observable)
    end
  
    [observable_links, incident_nodes]
  end

  def related_events(incident_id, observable_name)
    incident = Incident.find(incident_id)
    elements = get_incident_druid_data(incident, {})
    
    filtered_elements = elements.select do |element|
      element[:observables].any? do |observable|
        observable.key?(observable_name)
      end
    end

    filtered_elements
  end
end
