#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module IncidentsHelper
  def incidents_uniq_users(incidents, user = current_user) 
    incidents.includes(:users).map { |incident| incident.users.map(&:name) }.flatten.uniq || []
  end

  def incidents_uniq_sources(incidents, user = current_user)
    incidents.pluck(:source).uniq.reject(&:blank?) || []
  end

  def worklog_uniq_types(user = current_user)
    [['Notes','Notes'], ['Incident Changes', 'Incident Changes'], ['Response Actions', 'Response Actions']]
  end

  def worklog_uniq_users(worklog_data, user = current_user)
    filtered_worklog_data = worklog_data.reject { |entry| entry[:user].nil? || entry[:user].is_a?(String)}
    filtered_worklog_data.pluck(:user).uniq || []
  end

  def worklog_uniq_orders(user = current_user)
    [['Newest','Newest'], ['Oldest', 'Oldest']]
  end

  def priority_color(priority_name)
    case priority_name.downcase
    when "low"
      "#ffffff"
    when "medium"
      "#ffffff"
    when "high"
      "#ffffff"
    when "critical"
      "#ffffff"
    when "none"
      "#ffffff"
    when "unknow"
      "#ffffff"
    when "info"
      "#ffffff"
    else
      "#ffffff"
    end
  end

  def priority_bgcolor(priority_name)
    case priority_name.downcase
    when "low"
      "#89a54e"
    when "medium"
      "#c69958"
    when "high"
      "#aa4643"
    when "critical"
      "#ff4643"
    when "none"
      "#30BF93"
    when "unknow"
      "#919191"
    when "info"
      "#5eca5e"
    else
      "#000"
    end
  end

  def get_module_name_by_source(source)
    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:incident_source_name] && v[:incident_source_name] == source }

    incident_modules.first
  end

  def get_incident_filters_link(incident)
    return nil unless incident.source

    mod = get_module_name_by_source(incident.source)

    return nil unless mod

    module_name = mod.first

    hash = {}

    hash[0] = {
        "column" => "incident_uuid",
        "operator" => "is",
        "value" => incident.uuid,
        "enabled" => true
    }

    start_time = incident.created_at - 1.hour
    
    hash[1] = {
      "column" => "start_time",
      "operator" => "is",
      "value" => start_time,
      "enabled" => true
    }
    
    end_time = incident.created_at + 1.day
    hash[2] = {
      "column" => "end_time",
      "operator" => "is",
      "value" => end_time,
      "enabled" => true
    }

    json_string = hash.to_json
    encoded_json = CGI.escape(json_string)
      
    "/#{module_name}/tops?search=#{encoded_json}"
  end

  def get_detection_element_filters_link(element)
    hash = {}

    element[:url_filters].each_with_index do |e, index|
      field = e.first
      value = e.last

      hash[index] = {
        "column" => field,
        "operator" => "is",
        "value" => value,
        "enabled" => true
      }
    end

    json_string = hash.to_json
    encoded_json = CGI.escape(json_string)
    
    "/#{element[:module_name]}/tops?search=#{encoded_json}"
  end

  def get_incident_druid_data(incident, search_params, user_id=nil)
    # set user_id
    unless user_id  
      user_id = current_user ? current_user.id : User.first.id unless user_id 
    end

    data = []
    incident_modules = Redborder::Module.registered_modules.select { |_, v| v[:columns] && v[:columns][:incident_uuid] }
    incident_modules.each do |mod|
      source = Redborder::DruidSource.get(source: mod.first)
      filter = Redborder::Filter.new(source, {})
      filter.add(:incident_uuid, incident.uuid)

      start_time = incident.created_at - 1.hour
      end_time = start_time + 1.day

      # group_by: @source.default_columns_raw,
      data += source.where({
        query_type: :group_by,
        range: :custom,
        start_time: start_time,
        end_time: end_time,
        granularity: :minute,
        aggregation: source.default_aggregation,
        group_by: source.default_columns_incidents,
        search: filter.get,
        user: User.find_by(id: user_id)

      }).map do |t|
        url_filters = {}
        url_filters[:start_time] = start_time
        url_filters[:end_time] = end_time

        priority = 'N/A'
        if t.methods.include?(:priority)
          priority = t.send(:priority) || 'N/A'
        end

        sig_id = 'N/A'
        if t.methods.include?(:sig_id)
          sig_id = t.send(:sig_id) || 'N/A'
        end

        source_name = mod.last[:incident_source_name] rescue 'N/A'
        module_name = mod.first
        timestamp = t.send(:timestamp)
        msg = t.methods.include?(:msg) ? t.send(:msg) : nil

        # Observables are define in the columns of each module
        observables = []
        assets = []
        mod.last[:columns].select { |column, cfg| cfg[:observable_type] }.each do |column, cfg|
          asset = nil
          url_filters[column] = t.send(column)

          # We check if this is an asset
          if cfg[:object_type] && t.methods.include?(column) && t.send(column)
            asset = RedborderObject.find_by(value: t.send(column), type: cfg[:object_type], user: User.find_by(id: user_id))
          end

          if asset
            assets.push({ t.send(column) => cfg[:object_type] }) if t.methods.include?(column) && t.send(column)
          else
            observables.push( { t.send(column) => cfg[:observable_type] }) if t.methods.include?(column) && t.send(column)
          end
        end

        url_filters['incident_uuid'] = incident.uuid

        { timestamp: timestamp, sig_id: sig_id, priority: priority, msg: msg, source: source_name, assets: assets, observables: observables, module_name: module_name, url_filters: url_filters }
      end
    end

    # Filter by search
    if search_params && search_params[:source]
      data = data.select { |element| element[:source] == search_params[:source] }
    end

    if search_params && search_params[:priority]
      data = data.select { |element| element[:priority] == search_params[:priority] }
    end

    data
  end
end