#######################################################################
# Copyright (c) 2024 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module Redborder
  module Jobs
    # This jobs generates content: titles, description, etc. depending of the function
    class UpdateContentIncidentsJob < Struct.new(:verbose, :user_id)
      include Redborder::Jobs::JobHelper
      include IncidentsHelper

      def perform
        content = ''
        begin
          content += "Searching for non visible incidents..\n" if verbose
          incidents = Incident.all.where(visible: false)
          content += "#{incidents.count} incidents found.\n" if verbose

          incidents.each do |incident|
            content += "Getting data for incident #{incident.id}..\n" if verbose
            alerts = get_incident_druid_data(incident, {}, user_id)
            next unless alerts && !alerts.empty?

            params = { alerts: alerts }

            content += "Generating title for incident #{incident.id}\n" if verbose
            title = Redborder::AI.generate_incident_title(params)
            
            if title.is_a?(String)
              incident.name = title[0, 255]
              content += "Title for incident #{incident.id}: #{title}\n" if verbose
            else 
              content +="Error generating title for incident #{incident.id}\n" if verbose
            end

            description = Redborder::AI.generate_incident_description(params)
            if description.is_a?(String)
              incident.description = description[0, 1000]
              content += "Description for incident #{incident.id}: #{description}\n" if verbose
            else
              content += "Error generating description for incident #{incident.id}\n" if verbose
            end

            content += "Making incident #{incident.id} visible\n" if verbose
            incident.visible = true
            content += "Saving incident #{incident.id}\n" if verbose
            incident.save
            logit "Generated title and description for incident #{incident.id}\n" if verbose
          end
        rescue Exception => e
          content += "#{e.class}: #{e.message}\n"
          e.backtrace.each do |line|
            content += "#{line}\n"
          end
          logit "#{e.class}: #{e.message}"
        ensure
          StoredDelayedJob.create(job: "UpdateContentIncidentsJob", user_id: user_id, content: content)
          # Enqueue the job again
          Delayed::Job.enqueue(
          Redborder::Jobs::UpdateContentIncidentsJob.new(verbose, user_id), priority: 10, run_at: Time.now + 5.minutes)
        end
      end
    end
  end
end
